from typing import List, Optional

from serum import inject
import re

from .common.financial_symbol_id import FinancialSymbolId
from .common.financial_symbol import FinancialSymbol
from ._sources.registries import FinancialSymbolsRegistry
from ._sources.all_sources import AllSymbolSources


@inject
class _Search:
    sources: AllSymbolSources
    fin_syms_registry: FinancialSymbolsRegistry

    def __init__(self):
        self.id2sym = {}

        def handle_quandl_info(x, src):
            fin_sym = src.fetch_financial_symbol(x.fin_sym_id.name)

            line = '{} {} {}'.format(fin_sym.name, fin_sym.exchange, fin_sym.short_name)
            line = re.sub(r'\s+', ' ', line.lower())

            self.id2sym.update({line: fin_sym})

            return line

        def handle_micex_stocks(x, src):
            fin_sym = src.fetch_financial_symbol(x.fin_sym_id.name)

            line = '{} {} {} {}'.format(fin_sym.name, fin_sym.exchange, fin_sym.isin, fin_sym.long_name)
            line = re.sub(r'\s+', ' ', line.lower())

            self.id2sym.update({line: fin_sym})

            return line

        def handle_mutru(x, src):
            fin_sym = src.fetch_financial_symbol(x.fin_sym_id.name)

            line = '{} {}'.format(fin_sym.name, fin_sym.short_name)
            line = re.sub(r'\s+', ' ', line.lower())

            self.id2sym.update({line: fin_sym})

            return line

        quandl_lines = [handle_quandl_info(x, self.sources.quandl_source)
                        for x in self.sources.quandl_source.get_all_infos()]
        micex_stocks_lines = [handle_micex_stocks(x, self.sources.micex_stocks_source)
                              for x in self.sources.micex_stocks_source.get_all_infos()]
        mutru_lines = [handle_mutru(x, self.sources.mut_ru_source)
                       for x in self.sources.mut_ru_source.get_all_infos()]

        self.lines = quandl_lines + micex_stocks_lines + mutru_lines

    def _check_finsym_access(self, query: str) -> Optional[FinancialSymbol]:
        namespaces = self.fin_syms_registry.namespaces()

        starts_with_namespace = False
        for ns in namespaces:
            if query.startswith(ns + '/'):
                starts_with_namespace = True
                break
        if not starts_with_namespace:
            return None

        fsid = FinancialSymbolId.parse(query)
        return self.fin_syms_registry.get(fsid)

    @inject
    def perform(self, query: str, top: int) -> List[FinancialSymbol]:
        try:
            fs = self._check_finsym_access(query=query)
        except Exception:
            fs = None

        if fs:
            return [fs]
        else:
            if not isinstance(query, str):
                raise ValueError('`query` should be string')
            if not isinstance(top, int):
                raise ValueError('`top` should be int')

            top = max(0, top)
            if not query or top == 0:
                return []

            query = re.sub(r'\s+', ' ', query.strip().lower())
            if len(query) == 0:
                return []

            r = [(l.find(query), l) for l in self.lines]
            r = filter(lambda x: x[0] != -1, r)
            r = sorted(r, key=lambda x: '{:4d} {}'.format(x[0], x[1]))
            r = r[:top]
            r = [self.id2sym[x[1]] for x in r]
            return list(r)[:top]



from ..common.time_series import TimeSeries, TimeSeriesKind
from .._settings import _MONTHS_PER_YEAR

import numpy as np


def compute(cbr_top10_rates):
    index_total = None
    start_period = min(cbr_top10_rates['period'])
    end_period = max(cbr_top10_rates['period'])
    for month_idx in range(_MONTHS_PER_YEAR):
        rates_yearly = cbr_top10_rates['rate'].values[month_idx::_MONTHS_PER_YEAR]
        index_part = TimeSeries(values=np.repeat(rates_yearly, _MONTHS_PER_YEAR)[:len(cbr_top10_rates) - month_idx],
                                start_period=start_period + month_idx + 1,
                                end_period=end_period + 1,
                                kind=TimeSeriesKind.VALUES)
        index_part = (1 + index_part / _MONTHS_PER_YEAR).cumprod()
        index_total = index_part if index_total is None else index_total[1:] + index_part

    index_total = (index_total.pct_change() + 1.).cumprod() - 1.
    index_total = [0] + index_total
    index_total = (index_total + 1) * 100
    return index_total



from serum import singleton, inject
import pandas as pd

from .base_classes import FinancialSymbolsSource
from .single_financial_symbol_source import CbrTopRatesSource
from ..common.financial_symbol_info import FinancialSymbolInfo
from ..common.enums import Currency, SecurityType, Period
from ..common.financial_symbol_id import FinancialSymbolId
from ..common.financial_symbol import FinancialSymbol
from .._index.okid10 import compute as compute_top10


@singleton
@inject
class OkamaSource(FinancialSymbolsSource):
    cbr_top_rates_source: CbrTopRatesSource

    def __init__(self):
        super().__init__(namespace='index')

        self.cbr_top10_sym = self.cbr_top_rates_source.fetch_financial_symbol('TOP_rates')
        self.okid10 = 'OKID10'

    def fetch_financial_symbol(self, name: str):
        if name == self.okid10:
            def fetch_values(start_period, end_period):
                vals = self.cbr_top10_sym.values(
                    start_period=self.cbr_top10_sym.start_period,
                    end_period=end_period,
                )
                index_vals = compute_top10(vals)
                drop_first = max(0, (start_period - index_vals.start_period).n)
                index_vals = index_vals[drop_first:]
                df = pd.DataFrame({'date': pd.date_range(index_vals.start_period.to_timestamp(),
                                                         (index_vals.end_period + 1).to_timestamp(),
                                                         freq='M'),
                                   'period': pd.period_range(index_vals.start_period,
                                                             index_vals.end_period,
                                                             freq='M'),
                                   'close': index_vals.values})
                return df
            sym = FinancialSymbol(identifier=FinancialSymbolId(namespace=self.namespace, name=name),
                                  values=fetch_values,
                                  short_name='Okama TOP10 Index',
                                  long_name='Okama TOP10 Index',
                                  start_period=self.cbr_top10_sym.start_period,
                                  end_period=self.cbr_top10_sym.end_period,
                                  currency=Currency.RUB,
                                  security_type=SecurityType.INDEX,
                                  period=Period.MONTH,
                                  adjusted_close=True)
            return sym

    def get_all_infos(self):
        fsi = FinancialSymbolInfo(
            fin_sym_id=FinancialSymbolId(self.namespace, self.okid10),
            short_name=self.okid10,
        )
        infos = [fsi]
        return infos
